#include <stdio.h>
#include <stdlib.h>
#include "portaudio.h"
#include <gsl/gsl_fft.h>

#define SAMPLE_RATE  (8192)
#define NUM_SECONDS     (1)
#define NUM_CHANNELS    (1)
#define DITHER_FLAG     (0) 
#define FRAMES_PER_BUFFER  (1024)

#define THRESHOLD 10.0  // minimum "strength" to notice
#define PA_SAMPLE_TYPE  paFloat32
typedef float SAMPLE;
#define SAMPLE_SILENCE  (0.0f)

// We will hold 16 buffers of data in a ring
// so we can get behind a few buffers on processing
#define NRBUF 16  
typedef struct
{
  int         next;
  int         first;
  SAMPLE      *recordedSamples[NRBUF];
}
paTestData;

/* Called when touch tone detected */
// to avoid "runt errors" we need to see a digit twice in a row
// which means we need at least 1/4 second tone (this is not
// in spec for DTMF, by the way)
void detected(char digit)
{
  // make sure we get two of each digit and then reject
  // the rest (until we detect a gap)
  static int lastdigit=0;
  static int lastput=0;
  if (digit&&digit!=lastput&&digit==lastdigit) 
    {
      putchar(digit);   
      lastput=digit;
      fflush(stdout); 
    }
  if (!digit) lastput=0;
  lastdigit=digit;
}



/* This routine will be called by the PortAudio engine when audio is needed.
** It may be called at interrupt level on some machines so don't do anything
** that could mess up the system like calling malloc() or free().
*/
static int recordCallback( void *inputBuffer, void *outputBuffer,
                           unsigned long framesPerBuffer,
                           PaTimestamp outTime, void *userData )
{
  // Get data structure 
    paTestData *data = (paTestData*)userData;
    SAMPLE *rptr = (SAMPLE*)inputBuffer;
    SAMPLE *wptr = data->recordedSamples[data->next];
    long framesToRecord;
    long i;
    // pick up next buffer and rotate 
    if (++data->next==NRBUF) data->next=0;
    (void) outputBuffer; /* Prevent unused variable warnings. */
    (void) outTime;
    
    // Silence? 
    if( inputBuffer == NULL )
    {
        for( i=0; i<FRAMES_PER_BUFFER; i++ )
        {
            *wptr++ = SAMPLE_SILENCE;
        }
    }
    else  // no, real data
    {
        for( i=0; i<FRAMES_PER_BUFFER; i++ )
        {
            *wptr++ = *rptr++;
        }
    }

    return 0;  // don't terminate
}



// This is an FFT wrapper that uses the
// GNU Scientific Library (GSL) FFT function
void fft(unsigned N, SAMPLE datareal[], int binnums[], SAMPLE  bins[] )
{
  // GSL wants an array of doubles not floats
  static double *ary=NULL;
  if (!ary) ary=malloc(sizeof(double)*N);  // malloc first time through
  unsigned i;
  if (!ary) return;
  for (i=0;i<N;i++) ary[i]=datareal[i];
  // Do FFT
  gsl_fft_real_radix2_transform(ary,1,N);
  for (i=0;i<8;i++)
    {
      int bin=binnums[i];
      // find magnitude (N-bin) is imaginary part
      bins[i]=sqrt(ary[bin]*ary[bin]+ary[N-bin]*ary[N-bin]);
      // look at next bin too and average
      bin++;
      bins[i]+=sqrt(ary[bin]*ary[bin]+ary[N-bin]*ary[N-bin]);
      bins[i]/2; // average
    }

  // The above code only looks at the bins we ask (plus the next bin)
  // which is more efficient than looking at everything
  // If you want to copy all the outputs
  //  for (i=0;i<N/2;i++) 
  //    {
  //    data[i]=ary[i];
  //    dataimg[i]=ary[N-i];
  //    }
  // this only copied the meaningful half
  // don't worry about the rest 

  // don't free ary because we'll keep using it  
}

/*******************************************************************/
// Note that if you don't supply an argument, the code gets the
// default audio input device. If you supply a number, you get that
// device. Portaudio numbers devices 0, 1, 2,... If you have multiple
// sound devices, you can use tools provided with Portaudio to dump
// the sound card numbers, but the default should be your default
// soundcard
// second argument, if present is the threshold
int main(int argc, char *argv[])
{
    PortAudioStream *stream;
    PaError    err;
    paTestData data;
    int        i;
    int        totalFrames;
    int        numSamples;
    int        numBytes;
    float threshold=THRESHOLD;
    SAMPLE     max, average, val;
    // The bins that hold DTMF data
    int binnums[]= { 87, 96, 106, 117, 151, 167, 184, 204 };
    SAMPLE bins[8];
    char matrix[4][4]= { { '1', '2', '3', 'A' },
			 {'4', '5', '6',  'B' },
			 { '7', '8', '9', 'C' },
			 { '*', '0', '#', 'D' }};
    SAMPLE lowv,highv;
    int low,high;

    data.first=0;
    data.next=0;
    if (argc>=3) threshold=atof(argv[2]);  // custom threshold
    numSamples = FRAMES_PER_BUFFER*NUM_CHANNELS;
    numBytes = numSamples * sizeof(SAMPLE);
    for (i=0;i<NRBUF;i++) 
      {
	data.recordedSamples[i] = (SAMPLE *) malloc( numBytes );
      }
    if( data.recordedSamples == NULL )
    {
        printf("Could not allocate record array.\n");
        exit(1);
    }

    // start portaudio
    err = Pa_Initialize();
    if( err != paNoError ) goto error;

    /* Record some audio. -------------------------------------------- */
    err = Pa_OpenStream(
              &stream,
              argc==1?Pa_GetDefaultInputDeviceID():atoi(argv[1]),  // device
              NUM_CHANNELS,
              PA_SAMPLE_TYPE,
              NULL,
              paNoDevice,  // no playback required here
              0,
              PA_SAMPLE_TYPE,
              NULL,
              SAMPLE_RATE,
              FRAMES_PER_BUFFER,            /* frames per buffer */
              0,               /* number of buffers, if zero then use default minimum */
              0, /* paDitherOff, // flags */
              recordCallback,
              &data );
    if( err != paNoError ) goto error;

    err = Pa_StartStream( stream );
    if( err != paNoError ) goto error;
    printf("Listening! (threshold=%f)\n",threshold); 
    fflush(stdout);

    // This loop will run forever, processing audio data
    while( Pa_StreamActive( stream ) )
    {
      while (data.next==data.first) Pa_Sleep(1);
      fft(FRAMES_PER_BUFFER,data.recordedSamples[data.first],binnums,bins);
      if (++data.first==NRBUF) data.first=0;
      // So let's see what we got
      // each bin  is worth 8hz so
      // 697hz - bin 87   (all rounded down)
      // 770hz - bin 96
      // 852hz - bin 106
      // 941hz - bin 117
      // 1209hz - bin 151
      // 1336hz - bin 167
      // 1477hz - bin 184
      // 1633hz - bin 204
      // so taking the average of two adjacent bins will
      // give us a good idea
      low=high=-1;
      lowv=highv=-1.0;
      // Find magnitude of results we are interested in
      for (i=0;i<8;i++)
	{
	  if (i<4)
	    {
	    if (bins[i]>lowv && bins[i]>=THRESHOLD) 
	      {
		lowv=bins[i];
		low=i;
	      }
	    }
	  else
	    {
	      if (bins[i]>highv && bins[i]>=THRESHOLD) 
		{
		  highv=bins[i]; 
		  high=i-4; 
		}
	    }
	}
      // check to see that we found something
      // ought to check that highv and lowv are not too different
      if (low>=0 && high>=0)
	detected(matrix[low][high]);
      else
	detected(0);  // tell detector no input found




    }  // end processing loop
    
    // Should never get here because callback never terminates
    // so Pa_StreamActive is always true
    err = Pa_CloseStream( stream );
    if( err != paNoError ) goto error;
    free( data.recordedSamples );
    Pa_Terminate();
    return 0;

error:
    Pa_Terminate();
    fprintf( stderr, "An error occured while using the portaudio stream\n" );
    fprintf( stderr, "Error number: %d\n", err );
    fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
    return -1;
}
